static inline struct mem_loc_t merge (const struct mem_loc_t x, 
				      const struct mem_loc_t y) 
{
  struct mem_loc_t r;
  data_kno_t k = (~(x.val ^ y.val)) & x.known & y.known;
  r.val = k & (x.val | y.val);
  r.known = k;
  return r;
}

static inline data_t set_bit (data_t val, data_t bit)
{
  return val | (0x1 << bit);
}

static inline data_t clear_bit (data_t val, data_t bit)
{
  return val & ~((data_t)0x1 << bit);
}

static inline void clear_bit_addr (struct memory *m, 
				   data_addr_t addr, 
				   data_t bit)
{
  m->core[addr].known = set_bit (m->core[addr].known, bit);
  m->core[addr].val = clear_bit (m->core[addr].val, bit);
}

static inline data_addr_t io_to_mem (io_data_addr_t io)
{
  return (data_addr_t) (io + platform->io.offset);
}

static inline data_addr_t extended_io_to_mem (io_data_addr_t io)
{
  return (data_addr_t) (io + platform->extended_io.offset);
}

static inline void clear_status_bit (struct memory *m, 
				     status_bit_t bit)
{
  clear_bit_addr (m, io_to_mem (SREG_ADDR), bit);
}

static inline bool is_bit_set (data_t val,
			       data_t bit)
{
  return ((val & (0x1 << bit)) != 0);
}

static inline bool is_bit_clear (data_t val,
				 data_t bit)
{
  return ((val & (0x1 << bit)) == 0);
}

static inline data_addr_t reg_to_mem (reg_t reg)
{
  return (data_addr_t) reg;
}

static inline io_data_addr_t mem_to_io (data_addr_t addr)
{
  return (io_data_addr_t) ((unsigned int) addr - platform->regs.len);
}

static inline void copy_mem_byte (struct memory *m, 
				  data_addr_t src, 
				  data_addr_t dst)
{
  m->core[dst] = m->core[src];
}

static inline void kill_mem (struct memory *m, data_addr_t addr)
{
  m->core[addr].val = 0x0;
  m->core[addr].known = 0x0;
}

static inline void kill_reg (struct memory *m, reg_t reg)
{
  kill_mem (m, reg_to_mem (reg));
}

static inline void kill_dest (struct memory *m,
			      const struct decoder_info *dec)
{
  // FIXME -- assert this?
  assert (dec->dst_reg_set);
  kill_reg (m, dec->dst_reg);
}

static inline void kill_x (struct memory *m) {
  kill_reg (m, (reg_t)26);
  kill_reg (m, (reg_t)27);
}

static inline void kill_y (struct memory *m) {
  kill_reg (m, (reg_t)28);
  kill_reg (m, (reg_t)29);
}

static inline void kill_z (struct memory *m) {
  kill_reg (m, (reg_t)30);
  kill_reg (m, (reg_t)31);
}

static inline bool is_bit_known (const struct memory *m, 
				 data_addr_t addr, 
				 data_t bit)
{
  return ((m->core[addr].known & (0x1 << bit)) != 0);
}

static inline bool is_status_bit_known (const struct memory *m, 
					status_bit_t bit)
{
  return is_bit_known (m, io_to_mem (SREG_ADDR), bit);
}

static inline bool is_status_bit_set (const struct memory *m, 
				      status_bit_t bit)
{
  return is_bit_set (m->core[io_to_mem(SREG_ADDR)].val, bit);
}

#if !USE_BDD

static inline void set_val (struct memory *m, 
			    data_addr_t dst, 
			    data_val_t val)
{
  m->core[dst].known = 0xff;
  m->core[dst].val = val;
}

static inline bool get_bit (const struct memory *m, 
			    data_addr_t addr, 
			    data_t bit)
{
  return is_bit_set (m->core[addr].val, bit);
}

static inline bool get_status_bit (const struct memory *m, 
				   status_bit_t bit)
{
  return get_bit (m, io_to_mem (SREG_ADDR), bit);
}

#endif // !USE_BDD

static inline void set_bit_addr (struct memory *m, 
				 data_addr_t addr, 
				 data_t bit)
{
  m->core[addr].known = set_bit (m->core[addr].known, bit);
  m->core[addr].val = set_bit (m->core[addr].val, bit);
}

static inline void set_status_bit (struct memory *m, 
				   status_bit_t bit)
{
  set_bit_addr (m, io_to_mem (SREG_ADDR), bit);
}

static inline void assign_bit (struct memory *m, 
			       data_addr_t addr, 
			       data_t bit, 
			       bool val)
{
  if (val) {
    set_bit_addr (m, addr, bit);
  } else {
    clear_bit_addr (m, addr, bit);
  }
}

static inline void assign_status_bit (struct memory *m, 
				      status_bit_t bit, 
				      bool val)
{
  assign_bit (m, io_to_mem (SREG_ADDR), bit, val);
}

static inline void kill_bit (struct memory *m, 
			     data_addr_t addr, 
			     data_t bit)
{
  m->core[addr].known = clear_bit (m->core[addr].known, bit);
  m->core[addr].val = clear_bit (m->core[addr].val, bit);
}

static inline void kill_status_bit (struct memory *m, 
				    status_bit_t bit)
{
  kill_bit (m, io_to_mem (SREG_ADDR), bit);
}

static inline void copy_bit (struct memory *m,
			     data_addr_t src, data_t src_bit, 
			     data_addr_t dst, data_t dst_bit)
{
  if (is_bit_set (m->core[src].known, src_bit)) {
    if (is_bit_set (m->core[src].val, src_bit)) {
      set_bit_addr (m, dst, dst_bit);
    } else {
      clear_bit_addr (m, dst, dst_bit);
    }
  } else {
    kill_bit (m, dst, dst_bit);
  }
}

