// A really simple vector implementation and an iterator for it

#define NO_SIZE (-1)
#define INIT_ALLOC (4)

template<class Key> class vec_iterator;

template<class Key>
class vec {
 public:
  int size;
  int alloc_size;
  Key *elts;

  typedef vec_iterator<Key> iterator;

  // see comment in state_int_t::init
  void init() {
    size = 0;
    alloc_size = 0;
    elts = NULL;
  };

  // constructors
  vec<Key>() : size(0), alloc_size(0), elts(NULL) {};

  // copy constructor -- need this to avoid shallow copy
  vec<Key>(const vec<Key>& k) {
    size = k.size;
    alloc_size = k.alloc_size;
    if (k.elts) {
      elts = (Key *) xmalloc (k.alloc_size * sizeof (Key));
      memcpy (elts, k.elts, k.alloc_size * sizeof (Key));
    } else {
      elts = NULL;
    }
  }

  // overloaded assignment -- need this to avoid shallow copy
  const vec<Key>& operator= (const vec<Key>& k) {
    if (this != &k) {
      size = k.size;
      alloc_size = k.alloc_size;
      if (k.elts) {
	elts = (Key *) xmalloc (k.alloc_size * sizeof (Key));
	memcpy (elts, k.elts, k.alloc_size * sizeof (Key));
      } else {
	elts = NULL;
      }
    }
    return *this;
  };

  ~vec<Key>() {
    if (elts) free (elts);
  };

  bool is_bottom () const {
    return (size == NO_SIZE);
  };

  bool operator == (const class vec<Key> b) const {
    if (this->is_bottom()) {
      return b.is_bottom();
    }
    if (b.is_bottom()) {
      return false;
    }
    if (size != b.size) {
      return false;
    }
    for(int i=0; i<size; ++i) {
      if (elts[i] != b.elts[i]) {
        return false;
      }
    }
    return true;
  };

  bool operator != (const class vec<Key> b) const {
    return !(*this == b);
  };

  void push(Key a) {

    if (this->is_bottom()) return;

    assert (size <= alloc_size);
    assert ((alloc_size == 0 && !elts) ||
	    (alloc_size > 0 && elts));

    if (!elts) {
      alloc_size = INIT_ALLOC;
      elts = (Key *) xmalloc (INIT_ALLOC * sizeof (Key));
    }

    if (size==alloc_size) {
      alloc_size *= 2;
      // cout << "resizing a vector to " << alloc_size << "\n";
      elts = (Key *) xrealloc (elts, alloc_size * sizeof (Key));
    }

    elts[size] = a;
    size++;
  };

  void set_to_bottom() {
    size = NO_SIZE;
  }

  Key pop() {
    if (this->is_bottom() || size==0) return Key::bottom;
    size--;
    return elts[size];
  };

  iterator begin() const;
  iterator end()   const;

  //  static const vec<Key> bottom;
};

template<class Key>
ostream& operator << (ostream& os, const class vec<Key> b) {
  if (b.is_bottom()) {
    return os << "_|_";
  } else {
    os << "< ";
    for (int i=0; i<b.size; ++i) {
      os << b.elts[i] << endl << "  ";
    }
    os << '>';
    return os;
  }
};

template<class Key> 
class vec_iterator {
  int x;
  const vec<Key> &s;
 public:
  vec_iterator<Key>(const vec<Key> &_s, int _x) 
    : x(_x), s(_s) 
    {};
  Key operator*() {
    //    assert(0 <= x);
    assert(x < s.size);
    if (s.is_bottom()) return Key::bottom;
    return s.elts[x];
  };
  vec_iterator<Key> operator ++ () {
    ++x;
    return *this;
  };
  bool operator == (const class vec_iterator<Key> b) const {
    return (x == b.x);
  };
  bool operator != (const class vec_iterator<Key> b) const {
    return !(*this == b);
  };
};

template<class Key> 
inline vec_iterator<Key> vec<Key>::begin() const {
  assert( !(this->is_bottom()) );
  return vec_iterator<Key>(*this,0);
};

template<class Key> 
inline vec_iterator<Key> vec<Key>::end() const {
  return vec_iterator<Key>(*this,size);
};
