#include "stacktool.h"

#if WCET_ANALYSIS

////////////////////////////////////////////////////////////////
// Fixup link
////////////////////////////////////////////////////////////////

abs_loc fixup_link(abs_loc PC, code_ctxt_t ctx) {
  // It is my belief that I have stomped on all cases where a branch
  // had a broken context value.
  // But my belief is wrong so I'm commenting out this assert again.
  // ARGH!  
  //   assert (PC.ctx != NO_CTX);
  // The problem is hidden somewhere in the stacktool.  
  // Null abs_loc contexts are sometimes passed on when new ones 
  // should be created.

  if (PC.ctx == NO_CTX) {
    return abs_loc(PC.addr,ctx);
  } else {
    return PC;
  }
}

////////////////////////////////////////////////////////////////
// Utility
////////////////////////////////////////////////////////////////

// Disassemble up to but excluding some instruction
void disasm_range( code_addr_t start, code_addr_t end, FILE* outf, ostream &os)
{
  for( code_addr_t a = start; a < end; a=nextPC(a)) {
    os << "\t" << a << "\t";
    avr_disasm_insn(a,outf);
    os << endl;
  }
}

////////////////////////////////////////////////////////////////
// Read data from file
//
// File format is:
//
//    cost LABEL xxxx yyyy == please report cost from xxxx to yyyy
//    count HHHH DDDD      == number of executions of 0xHHHH is DDDD 
//    Cost LABEL loc1 loc2 == please report cost from srcloc1 to srcloc2
//    Count srcloc DDDD    == number of executions of srcloc is DDDD 
//
// Only counts implemented so far
////////////////////////////////////////////////////////////////

class cost_range {
 public: 
  char* label;
  code_addr_t start, end;
  cost_range() {};
  cost_range(const char* _label, int n, code_addr_t _start, code_addr_t _end) 
    : start(_start), end(_end) {
    label = (char*)xmalloc(n+1);
    strncpy(label,_label,n);
  };
};

class count {
 public: 
  srcloc loc;
  int cnt;
  count() {};
  count(const srcloc &_loc, int _count) : loc(_loc), cnt(_count) {};
};

// count entry using absolute address instead of srcloc
class abs_count {
 public: 
  code_addr_t addr;
  int cnt;
  abs_count() {};
  abs_count(const code_addr_t &_addr, int _count) : addr(_addr), cnt(_count) {};
};

static queue<cost_range> cost_ranges;
static queue<srcloc>     unknown;
static queue<count>      counts;
static queue<count>      unrolls;
static queue<abs_count>  abs_counts;

static const int MAX_LINE_LENGTH = 1000;

bool read_addr(char* &s, code_addr_t *m) {
  return read_num(s,(int*)m,16);
}

void read_wcet_from_file( const char* fn ) {
  FILE* f = fopen(fn,"r");
  if (f == NULL) {
    cerr << "Error: input file '" << fn << "' cannot be opened for reading." 
         << endl;
    die();
  }

  char buffer[MAX_LINE_LENGTH];
  int line = 0;
  do {
    char* b = fgets(buffer,MAX_LINE_LENGTH,f);
    line++;
    char* s = b;
    code_addr_t m,n;
    int i;
    char* l;
    int len;
    srcloc loc1,loc2;
    if (s == NULL) {
      return;

    } else if (s=b, matches(s,(char*)"Prefix") 
        && skip_spaces(s)  && read_token(s,&l,&len)
        && skip_spaces0(s) && *s == '\0'
        ) {
      srcloc_prefixes.push(string_copy(l,len));

    } else if (s=b, matches(s,(char*)"ICall") 
        && skip_spaces(s)  && read_token(s,&l,&len)
        && skip_spaces0(s) && *s == '\0'
        ) {
      icall_funs.push(string_copy(l,len));

    } else if (s=b, matches(s,(char*)"count") 
        && skip_spaces(s) && read_addr(s,&m)
        && skip_spaces(s) && read_num(s,&i,10)
        && skip_spaces0(s) && *s == '\0'
        ) {
      abs_counts.push(abs_count(m,i));

    } else if (s=b, matches(s,(char*)"UCount") 
        && skip_spaces(s) && read_srcloc(s,&loc1)
        ) {
      unknown.push(loc1);

    } else if (s=b, matches(s,(char*)"Count") 
        && skip_spaces(s) && read_srcloc(s,&loc1)
        && skip_spaces(s) && read_num(s,&i,10)
        && skip_spaces0(s) && *s == '\0'
        ) {
      counts.push(count(loc1,i));

    } else if (s=b, matches(s,(char*)"Unroll") 
        && skip_spaces(s) && read_srcloc(s,&loc1)
        && skip_spaces(s) && read_num(s,&i,10)
        && skip_spaces0(s) && *s == '\0'
        ) {
      unrolls.push(count(drop_srcloc_prefix(loc1),i));

    } else if (s=b, matches(s,(char*)"cost") 
        && skip_spaces(s) && read_token(s,&l,&len)
        && skip_spaces(s) && read_addr(s,&m)
        && skip_spaces(s) && read_addr(s,&n)
        && skip_spaces0(s) && *s == '\0'
        ) {
      cost_ranges.push(cost_range(l,len,(code_addr_t)m,(code_addr_t)n));
#if 0
      if (state(abs_loc(m,0))->head != state(abs_loc(n,0))->head) {
        cerr << "Addresses " << m << " and " << n 
             << " are in different functions.  Oh well."
             << endl;
      }
#endif
    } else if (s=b, matches(s,(char*)"Cost") 
        && skip_spaces(s) && read_token(s,&l,&len)
        && skip_spaces(s) && read_srcloc(s,&loc1)
        && skip_spaces(s) && read_srcloc(s,&loc2)
        && skip_spaces0(s) && *s == '\0'
        ) {
      cost_ranges.push(cost_range(l,len,(code_addr_t)m,(code_addr_t)n));
#if 0
      // ToDo: if a function is inlined multiple times, we should
      // locate all regions bracketed by these srclocs.
      m = find_first_srcloc(loc1);
      n = find_last_srcloc(loc2);
      if (m == NO_CODE_ADDR && n == NO_CODE_ADDR) {
        cout << "Ignoring Cost request " << b;
        continue;
      }
      assert( m != NO_CODE_ADDR );
      assert( n != NO_CODE_ADDR );
      decode_addr(m);  // make sure we've allocated instructions
      decode_addr(n); 
      assert( all_code[m]->state_cnt);
      assert( all_code[n]->state_cnt);
      if (state(abs_loc(m,0))->head != state(abs_loc(n,0))->head) {
        cerr << "Addresses " << m << " and " << n 
             << " are in different functions.  Oh well."
             << endl;
      }
#endif
    } else if (s=b, skip_spaces0(s) && *s == '\0') { // blank lines
    } else if (s=b, *s == '#') { // comments
    } else {
      cerr << "Parse error in file " << fn << ":" << dec << line << hex << endl;
      die(); // parse error
    }
  } while (1);
}

////////////////////////////////////////////////////////////////
// Looking up loops in the loop count table
////////////////////////////////////////////////////////////////

static int lookup_count(queue<count> is, srcloc loc) {
  for(queue<count>::iterator i = is.begin(); i != is.end(); ++i) {
    if ((*i).loc == loc) {
      return (*i).cnt;
    }
  }
  return -1;
}

// look up unroll counts read from file
static int lookup_unroll_count(code_addr_t from, code_addr_t to) {

  // Try source locs
  srcloc loc;
  int c = 1;  // raw count
  if (addr_to_srcloc(from,&loc)) {
    c = max<int>(c,lookup_count(unrolls,loc));
  }
  if (addr_to_srcloc(to,&loc)) {
    c = max<int>(c,lookup_count(unrolls,loc));
  }
  return c;
}
    
// look up loop counts read from file
static int lookup_loop_count(code_addr_t from, code_addr_t to) {

  // absolute loop counts have priority
  for(queue<abs_count>::iterator i = abs_counts.begin(); i != abs_counts.end(); ++i) {
    if ((*i).addr == from) {
      return (*i).cnt;
    }
  }

  // Try source locs
  srcloc loc;
  int c = -1;  // raw count
  if (addr_to_srcloc(from,&loc)) {
    if (c == -1) c = lookup_count(counts,loc);
  }
  if (addr_to_srcloc(to,&loc)) {
    if (c == -1) c = lookup_count(counts,loc);
  }
  if (c != -1) {
    return c;
  }
    
  if (avr_is_jmp(from) && from == to) {
    cout << "# Warning: Infinite loop at " << from << ".." << to << endl;
    if (1 || debug_wcet>2) {
      cout << "Loop body:" << endl;
      disasm_range(from,to+1,stdout,cout);
    }
    return 1000000;
  }

  // Report an expected failure
  if (addr_to_srcloc(from,&loc)) {
    for(queue<srcloc>::iterator i = unknown.begin(); i != unknown.end(); ++i) {
      if ((*i) == loc) {
        cout << "# Warning: Missing loop count for " << loc << ".  Analysis will be unsound." << endl;
        return NEG_INFINITY;
      }
    }
  }
  if (addr_to_srcloc(to,&loc)) {
    for(queue<srcloc>::iterator i = unknown.begin(); i != unknown.end(); ++i) {
      if ((*i) == loc) {
        cout << "# Warning: Missing loop count for " << loc << ".  Analysis will be unsound." << endl;
        return NEG_INFINITY;
      }
    }
  }

  // Report failure
  cout << "# Warning: Missing loop count for loop " 
       << from << ".." << to << ".  Analysis will be unsound." << endl;
  queue<srcloc> *locs = 0;
  if (addr_to_srclocs(from,&locs)) {
    for(queue<srcloc>::iterator i = locs->begin(); i != locs->end(); ++i) {
      cout << "#  Count " << *i << " ???" << endl;
    }
  }
  if (addr_to_srclocs(to,&locs)) {
    for(queue<srcloc>::iterator i = locs->begin(); i != locs->end(); ++i) {
      cout << "#  Count " << *i << " ???" << endl;
    }
  }
  cout << "#  count " << from << " ???" << endl;
  if (debug_wcet>2) {
    cout << "Loop body:" << endl;
    disasm_range(from,to+1,stdout,cout);
  }
  return NEG_INFINITY;

}

////////////////////////////////////////////////////////////////
// An edge between basic blocks
// (often used for retreating edges)
////////////////////////////////////////////////////////////////

class edge {
 public:
  code_addr_t source;
  code_addr_t target;
  edge() {};
  edge(code_addr_t _s, code_addr_t _t) : source(_s), target(_t) {};

  bool operator == (const class edge e) const {
    return (source == e.source && target == e.target);
  };
  bool operator != (const class edge e) const {
    return !(*this == e);
  };
};

extern ostream& operator << (ostream& os, const class edge &e);

ostream& operator << (ostream& os, const class edge &e) {
  return os << e.source << "->" << e.target;
}

////////////////////////////////////////////////////////////////
// Loops - one or more backwards edges and a set of basic blocks
////////////////////////////////////////////////////////////////

// Loops are formed by merging overlapping, non-nested loops
class loop {
 public:
  queue<edge> retreating_edges; // retreating edges of loops
  queue<code_addr_t> blocks;    // basic blocks in this loop
  int count;                // execution count
  
  loop() {};
  loop(const edge &e, const queue<code_addr_t> &bs) {
    // exn_count = NEG_INFINITY;
    retreating_edges.insert(e);
    union2<code_addr_t>(blocks,bs);
  };

  void merge(const class loop &l) {
    union2<edge>(retreating_edges,l.retreating_edges);
    union2<code_addr_t>(blocks,l.blocks);
  };

  // Loops are ordered by inclusion of set of blocks
  bool operator <= (const class loop l) const {
    return (blocks <= l.blocks);
  };

  int size() const { return blocks.length(); }
  
};

ostream& operator << (ostream& os, const class loop &l) {
  return os << l.retreating_edges << "{ " << l.blocks << "}";
}

bool disjoint(const loop &x, const loop &y) {
  return disjoint<code_addr_t>(x.blocks,y.blocks);
};

////////////////////////////////////////////////////////////////
// Flowgraphs
////////////////////////////////////////////////////////////////

class flowgraph {
 public:
  queue<code_addr_t> blocks;
  queue<loop*> members;
  flowgraph() {};

  void insert (loop *l) {
    for(queue<loop*>::iterator i = members.begin(); i != members.end(); ++i) {
      if (!disjoint(*(*i),*l)) {
        if ((*i)->size() < l->size() && **i <= *l) {
          if (debug_wcet > 3) {
            cout << "  > " << *l << endl
                 << "    " << **i << endl;
          }
        } else if (l->size() < (*i)->size() && *l <= **i) {
          if (debug_wcet > 3) {
            cout << "  < " << *l << endl
                 << "    " << **i << endl;
          }
        } else {
          if (debug_wcet > 3) {
            cout << "  merging " << *l << endl
                 << "          " << **i << endl;
          }
          (*i)->merge(*l);
          return;
        }
      }
    }
    // Doesn't overlap with any existing loops - hurrah!
    if (debug_wcet > 3) {
      cout << "  inserting " << *l << endl;
    }
    members.push(l);
  };

};

ostream& operator << (ostream& os, const class flowgraph &ls) {
  return pp_list(os, ls.members, "\n  ");
}

////////////////////////////////////////////////////////////////
// Basic blocks
//
// We use the first instruction in a block as the label for a block.
// For the purposes of loop detection, we treat function calls as nops.
//
// Basic blocks make up the nodes that we will run the dfs on to 
// determine loops.  
//
////////////////////////////////////////////////////////////////

// Find the last instruction in a basic block.
code_addr_t to_block_end( code_addr_t a )
{
  while (all_code[a]->control.edges_out.len == 1
         && !all_code[a]->control.edges_out.list[0].dead) {

    // Treat call instructions specially - assume they return!
    if (avr_is_call(a)) { 
      code_addr_t n = nextPC(a);
      if (all_code[n] 
          && all_code[n]->control.edges_in.len == 0) {
        a = n;
      } else {
        return a;
      }
    } else if (avr_is_jmp(a)) { 
      return a;
    } else { // normal instructions
      code_addr_t n = all_code[a]->control.edges_out.list[0].target.addr;
      if (all_code[n] 
          && all_code[n]->control.edges_in.len == 1) {
        a = n;
      } else {
        return a;
      }
    }
  }
  return a;
}

// find start of basic block
code_addr_t to_block_start( code_addr_t a )
{
 restart:
  while (all_code[a]->control.edges_in.len == 1
         && !all_code[a]->control.edges_in.list[0].dead) {

    code_addr_t p = all_code[a]->control.edges_in.list[0].target.addr;

    if (avr_is_jmp(p) || avr_is_call(p)) { 
      return a;
    } else { // normal instructions
      if (all_code[p] 
          && all_code[p]->control.edges_out.len == 1) {
        a = p;
      } else {
        return a;
      }
    }
  }
  if (all_code[a] 
      && a > 0
      && all_code[a]->control.edges_in.len == 0) {
    code_addr_t prev = prevPC(a);
    if (avr_is_call(prev)) {
      a = prev;
      goto restart;
    }
  }
  return a;
}

////////////////////////////////////////////////////////////////
// Depth first traversal of basic blocks
////////////////////////////////////////////////////////////////

int dfn(code_addr_t a) {
  return all_code[a]->dfn;
}

// Based on Dragon book, algorithm 10.14, page 662
static int depth;

void dfs( queue<code_addr_t> &order, queue<edge> &rets,code_addr_t n) {
  assert(all_code[n]);

  all_code[n]->visited = true;

  code_addr_t m = to_block_end(n);
  if (avr_is_call(m)) { 
    // pretend that all calls return
    dfs(order,rets,nextPC(m));
  } else {
    for(unsigned int i=0; i<all_code[m]->control.edges_out.len; ++i) {
      struct branch_elt *b = &all_code[m]->control.edges_out.list[i];
      code_addr_t t = b->target.addr;
      if (!b->dead && !all_code[t]->visited) {
        dfs(order,rets,t);
      }
    }
  }

  all_code[n]->dfn = depth--;
  order.unshift(n);

  // Enumerate retreating edges
  if (avr_is_call(m)) { // skip
  } else {
    for(unsigned int i=0; i<all_code[m]->control.edges_out.len; ++i) {
      struct branch_elt *b = &all_code[m]->control.edges_out.list[i];
      code_addr_t t = b->target.addr;
      if (!b->dead && dfn(n) >= dfn(t)) {
        rets.push(edge(n,t));
      }
    }
  }

}

// reachability (forwards) - ignoring retreating edges
// we aren't able to use this
void reach_f(queue<code_addr_t> &loop, code_addr_t n) {
  
  if (!all_code[n]->visited) {
    all_code[n]->visited = true;
    loop.push(n);
    
    code_addr_t m = to_block_end(n);
    if (avr_is_call(m)) { 
      // pretend that all calls return
      reach_f(loop,nextPC(m));
    } else {
      for(unsigned int i=0; i<all_code[m]->control.edges_out.len; ++i) {
        struct branch_elt *b = &all_code[m]->control.edges_out.list[i];
        code_addr_t t = b->target.addr;
        if (!b->dead) {
          if (dfn(n) < dfn(t)) {
            reach_f(loop,t);
          }
        }
      }
    }
  } else {
    //    cout << "Ignoring " << n << endl;
  }
}


// reachability (backwards) - including retreating edges
// we want to start with the last instruction in the loop 
// and go backwards from there.  
void reach_b(queue<code_addr_t> &loop, code_addr_t n) {
  
  if (!all_code[n]->visited) {
    // cout << "    Marking " << n << endl;
    all_code[n]->visited = true;
    loop.push(n);
    
    if (all_code[n]->control.edges_in.len == 0
        && n > 0) {
      // check if it is a return site
      code_addr_t prev = prevPC(n);
      if (avr_is_call(prev)) { // pretend that all calls return
        reach_b(loop,to_block_start(prev));
      }
    } else {
      for(unsigned int i=0; i<all_code[n]->control.edges_in.len; ++i) {
        struct branch_elt *b = &all_code[n]->control.edges_in.list[i];
        code_addr_t t = b->target.addr;
        if (!b->dead && !avr_is_call(t)) {
          //          cout << "    Following " << t << endl;
          reach_b(loop,to_block_start(t));
        }
      }
      code_addr_t prev = prevPC(n);
      if (all_code[prev] && avr_is_call(prev)) {
        reach_b(loop,to_block_start(prev));
      }
    }
  } else {
    // cout << "    Ignoring " << n << endl;
  }
}

void print_basic_blocks( ostream &os, queue<code_addr_t> bs) {
  for(queue<code_addr_t>::iterator i = bs.begin(); i != bs.end(); ++i) {
    code_addr_t a = *i;
    os << "  Basic block " << a 
       << ".." << to_block_end(a)
       << ": " << dec << dfn(a) << hex
       << endl;
  }
}

void clear_visit_flags(void) {
  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    if (all_code[addr]) {
      all_code[addr]->visited = false;
    }
  }
}

////////////////////////////////////////////////////////////////
// Loop detection
//
// We detect loops as follows:
// 1) Find all retreating edges.
// 2) For each retreating edge e:
//    2.1) Delete all retreating edges except e
//    2.2) loop(e) = strongly connected component including e
//                 = edges reachable from end of loop traversing graph
//                   backwards and not visiting any blocks before
//                   the start of the loop
// 3) Merge loops which overlap
//
////////////////////////////////////////////////////////////////

void find_loops_in_function(flowgraph &ls, code_addr_t start) {
  if (debug_wcet > 3) {
    cout << addr_to_head_label(start) << endl;
  }
  // DFN numbers are allocated in decreasing order.  We could start
  // with a depth of 0 but I thought that would look funny so I decided
  // to start from 1000.  Use any number you like here.
  // (In most DFS literature, nodes are numberd from 1 to N where N
  // is the number of nodes but it'd take another pass to find what
  // N is first so we don't.)
  depth = 1000;
  queue<edge> rets;
  clear_visit_flags();
  dfs(ls.blocks,rets,start);
  //  print_basic_blocks(ls.blocks);
  for(queue<edge>::iterator i = rets.begin();
      i != rets.end();
      ++i) {
    queue<code_addr_t> members;

    // Mark as visited all subsequent nodes
    clear_visit_flags();
    for(queue<code_addr_t>::iterator k = ls.blocks.begin(); k != ls.blocks.end(); ++k){
      code_addr_t a = *k;
      if (dfn((*i).target) > dfn(a)) {
        all_code[a]->visited = true;
      }
    }
    reach_b(members,(*i).source);
    ls.insert(new loop(*i,members));
  }
}

// extract optimization info (number of copies made) from srcloc
static int get_divisor( const srcloc &l ) {
  const char* s = l.filename;
  int divisor = 1;
  int last_opt_id = -1;
  while ((s = strchr(s,'|')) != NULL) {
    ++s;
    int opt_id = -1;
    int div = -1;
    if (sscanf(s,"divide %d %d",&opt_id,&div) == 2) {
      if (opt_id != last_opt_id) {
        last_opt_id = opt_id;
        divisor *= div;
      }
    }
  }
  return divisor;
}

void analyze_function(code_addr_t addr) {
  flowgraph ls;
  find_loops_in_function(ls,addr);

  for(queue<loop*>::iterator l = ls.members.begin(); l != ls.members.end(); ++l){
    
    int count = NEG_INFINITY;
    int unrolling = NEG_INFINITY;
    for(queue<edge>::iterator j = (*l)->retreating_edges.begin(); 
        j != (*l)->retreating_edges.end(); ++j){
      count = max<int>(count,
                       lookup_loop_count((*j).target,
                                         to_block_end((*j).source)));
#if 0
      unrolling = max<int>(unrolling,
                           lookup_unroll_count((*j).target,
                                             to_block_end((*j).source)));
#else
      code_addr_t head = (*j).target;
      code_addr_t tail = to_block_end(head);
      for(code_addr_t i = head; i <= tail; i = nextPC(i)) {
        queue<srcloc> *locs;
        if (addr_to_srclocs(i,&locs)) {
          for(queue<srcloc>::iterator k = locs->begin(); k != locs->end(); ++k) {
            int d = get_divisor(*k);
            if (d > 1) {
              if (unrolling == NEG_INFINITY) {
                unrolling = d;
              } else if (unrolling != d) { 
                cout << "Whoa, dude: found two different divisors "
                     << dec << unrolling << " and " << d << hex
                     << " for loop " << *l
                     << endl;
              }
            }
          }
        }
      }
#endif
    }
    if (count == NEG_INFINITY) {
      count = 10;
    }
    if (debug_wcet > 0) {
      if (unrolling != NEG_INFINITY) {
        cout << "Using unroll count " << dec << unrolling << hex 
             << " for loop " << **l << endl;
      } else {
        cout << "No unroll count for loop " << **l << endl;
      }
    }
    if (unrolling != NEG_INFINITY) {
      count /= unrolling;
    }
    if (debug_wcet > 0) {
      cout << "Using loop count " << dec << count << hex 
           << " for loop " << **l << endl;
    }
    (*l)->count=count;
  }
  
  for(queue<code_addr_t>::iterator block = ls.blocks.begin(); 
      block != ls.blocks.end(); ++block){
    code_addr_t head = *block;
    code_addr_t tail = to_block_end(head);

    if (debug_wcet > 2) {
      cout << " DFN(" << head << ") = " << dec << dfn(head) << hex << endl;
    }
    int count = 1;
    for(queue<loop*>::iterator l = ls.members.begin(); 
        l != ls.members.end(); ++l){
      if ((*l)->blocks.contains(head)) {
        if ((*l)->count != NEG_INFINITY) {
          count *= (*l)->count;
        }
      }
    }

    for(code_addr_t i = head; i <= tail; i = nextPC(i)) {
      all_code[i]->exn_count = count;
      if (debug_wcet > 2) {
        cout << "  Setting cost of " << i << " to " << dec << count << hex << endl;
      }
    }
  }
}

// Scan entire program for backward branches, infer loop bounds 
// then lookup loop counts from file
void find_loops() {

#if 0
  for(code_addr_t addr = (code_addr_t) code_size-1; addr > 0 ; addr--) {
    if (all_code[addr]) {
      code_addr_t end = to_block_start(addr);
      cout << "Basic block from " << addr 
           << " to " << end
           << endl;
      if (end >= 0xc6) {
        disasm_range(end,addr+1,stdout,cout);
      }
      addr = min<int>(addr,end);
    }
  }
  cout << flush;
  exit(100);
#endif

  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    bool called = false;
    for (code_ctxt_t ctx=(code_ctxt_t)0; all_code[addr] && ctx < (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {
      abs_loc PC = abs_loc(addr,ctx);
      called |= (state(PC)->num_calls > 0);
    }
    if (called) {
      analyze_function(addr);
    }
  }
  for (unsigned int vec=0; vec < platform->num_vectors; vec++) {
    analyze_function(vec * VECTOR_SIZE);
  }
}

////////////////////////////////////////////////////////////////
// Scrub wcets from all states ready to start analysis
////////////////////////////////////////////////////////////////

void scrub_wcets() {
  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    if (all_code[addr]) {
      for(unsigned ctx = 0; ctx < all_code[addr]->state_cnt; ++ctx) {
        abs_loc PC = abs_loc(addr,ctx);
        state(PC)->wcet = NEG_INFINITY;
      }
    }
  }
}

////////////////////////////////////////////////////////////////
// WCET analysis
////////////////////////////////////////////////////////////////

// work queue (stored in wcet_queue)
// can be considered a set except for when analyzing the 
// function callgraph

class wcet_elt {
 public:
  abs_loc_t PC;
  cost_t wcet;
  wcet_elt() {};
  wcet_elt( abs_loc _PC, cost_t _wcet ) 
    : PC(_PC), wcet(_wcet) {};

  // only a partial match asked for!
  bool operator == (const wcet_elt &x) const {
    return (PC == x.PC);
  };
  bool operator != (const wcet_elt &x) const {
    return !(*this == x);
  };
};

static queue<wcet_elt> wcet_queue;

static void  wcet1(struct wcet_elt &e) {
  abs_loc PC = e.PC;
  if (debug_wcet > 3) {
    cout << "WCET analysis: " << PC 
         << " [" << dec << e.wcet << hex << "]" << endl;
  }
  assert(PC.addr != NO_CODE_ADDR);
  assert(PC.ctx != NO_CTX);

  // test that this instruction is even slightly live
  if (!is_valid_state(PC)) {
    if (debug_wcet > 2) {
      cout << "Skipping dead instruction " << PC << endl;
    }
    return;
  }

  if (all_code[PC.addr]->exn_count == NEG_INFINITY) { cout << PC << endl; }
  assert(all_code[PC.addr]->exn_count != NEG_INFINITY);

  // check that wcet is higher
  if (e.wcet <= state(PC)->wcet) return;

  // Got a higher number - propagate it
  state(PC)->wcet = e.wcet;

  // Propagate to any functions which may return to here
  //
  // (There is no edge from a call instruction to the return address
  // so, in most cases, only this loop or the next loop will execute.)

  // max cost of all functions returning to here
  cost_t fun_cost = NEG_INFINITY;

  for(int i=0; i<state(PC)->num_ret_targets; ++i) {
    abs_loc ret = state(PC)->ret_targets[i];
    abs_loc caller = prevPC(PC);
    bool pending = true;
    assert(avr_is_call(caller));

    if (state(caller)->terminate_analysis) {
      // Skip (use zero cost for) calls to bottomed out recursion
      // This is not entirely accurate.
      fun_cost = (cost_t)max<int>((cost_t)0, (int)fun_cost);
      if (debug_wcet > 3) {
        cout << "Skipping recursive call " 
             << caller << " -> " << PC
             << endl;
      }
    } else {
      // Propagate cost from fn head to here.
      // The problem is that we don't have a reliable set of heads
      // so we use the list of successors of the call instruction.
      // This is doing a depth first search of the callgraph and
      // then propagating them.  
      if (debug_wcet > 3) {
        cout << "  Chasing " << dec << state(caller)->out_edges.len << hex 
             << " heads for return edge " 
             << caller << "/" << ret
             << endl;
      }
      for(unsigned int j=0; j<state(caller)->out_edges.len; ++j) {
        struct branch_elt *b = &state(caller)->out_edges.list[j];
        if (b->dead) {
          cout << "  Ignoring dead call instruction at " << caller << endl;
        } else {
          abs_loc head = fixup_link(b->target,PC.ctx);
          if (debug_wcet > 3) {
            if (head != state(ret)->head) {
              cout << "  Found an extra head " << head << " for call site " 
                   << caller << "/" << ret
                   << endl;
            }
          }
          if (state(head)->wcet != NEG_INFINITY) {
            // Use cached value + return cost
            fun_cost = (cost_t)max<int>(state(head)->wcet + (cost_t)5,
                                        (int)fun_cost);
            if (debug_wcet > 3) {
              cout << "  Using cached value " 
                   << dec << state(head)->wcet << hex
                   << " for function " << head 
                   << " while analyzing " << PC << endl;
            }
          } else {
            pending = true;
            if (debug_wcet > 3) {
              cout << "  No cached value available for " << head << endl;
            }
          }
        }
      }
    }
    if (pending) {
      wcet_queue.push(wcet_elt(ret,(cost_t)5));
    }
    if (fun_cost != NEG_INFINITY) {
      // if some costs resolved, propagate them
      cost_t next_wcet = e.wcet + all_code[PC.addr]->exn_count * fun_cost;
      wcet_queue.push(wcet_elt(caller,next_wcet));
    }
  }

  // Propagate to any instructions that call this instruction
  for(int i=0; i<state(PC)->num_calls; ++i) {
    struct call_addr_ctx *b = &state(PC)->calls[i];
    abs_loc caller = b->call;
    abs_loc ret = b->ret;

    assert(is_valid_state(caller));
    assert(is_valid_state(ret));

    if (state(ret)->wcet != NEG_INFINITY) {
      cost_t next_wcet = state(ret)->wcet
                       + all_code[caller.addr]->exn_count * (e.wcet + 5);
      wcet_queue.push(wcet_elt(caller,next_wcet));
      if (debug_wcet > 3) {
        cout << "  Propagating cost " << dec << next_wcet << hex
             << " to " << caller << " from " << PC
             << endl;
      }
    } else {
      if (debug_wcet > 3) {
        cout << "  Ignoring unresolved fn call at " << caller << endl;
      }
    }
  }

  // Propagate upstream to any non-call/non-return instructions
  for(unsigned int i=0; i<state(PC)->in_edges.len; ++i) {
    struct branch_elt *b = &state(PC)->in_edges.list[i];
    // 2nd half of test protects against unconditional branch
    // from dead code
    if (!b->dead && all_code[b->target.addr]->state_cnt>0) {
      abs_loc target = fixup_link(b->target,PC.ctx);

      if (avr_is_call(target)) {
        // Nothing to do: we handled this above.
        if (debug_wcet > 3) {
          cout << "  Finished analyzing function " 
               << PC << " (called from " << target << ")."
               << " Cost = " << dec << e.wcet << hex
               << endl;
        }
      } else if (target.addr < PC.addr
                 || (all_code[PC.addr]->decode.label
                     && strcmp( all_code[PC.addr]->decode.label,
                                "__divmodhi4_exit")==0)) {
        if (is_valid_state(target)) {
          cost_t next_wcet = e.wcet + b->cost * all_code[target.addr]->exn_count;
          wcet_queue.push(wcet_elt(target,next_wcet));
          if (debug_wcet > 3) {
            cout << "Queueing " << target << endl;
          }
        }
      } else {
        if (debug_wcet > 2) {
          cout << "  Skipping backwards branch " 
               << target << "->" << PC
               << endl;
        }
      }
    }
  }
}

void wcet_fixpoint() {
  wcet_elt q;
  while (wcet_queue.pop(&q)) {
    wcet1(q);
  }
}

////////////////////////////////////////////////////////////////
// Cost of each function and irq vector
////////////////////////////////////////////////////////////////

void report_function_costs(void) {
  scrub_wcets();

  // Trace all live code starting from all return instructions
  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    if (avr_is_ret(addr)) {
      for (code_ctxt_t ctx=(code_ctxt_t)0; ctx < (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {
        abs_loc PC = abs_loc(addr,ctx);
        if (debug_wcet > 1) {
          cout << "Pushing return instruction " << PC << endl;
        }
        wcet_queue.push(wcet_elt(PC,(cost_t)5));
      }
    }
  }      
  wcet_fixpoint();
  for (unsigned int vec=0; vec < platform->num_vectors; vec++) {
    cost_t cost = state(abs_loc(vec * VECTOR_SIZE,0))->wcet;
    if (cost != NEG_INFINITY) {
      cout << "Cost of __vector_" << dec << vec << " = " << cost << hex
           << endl;
    }
  }
  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    for (code_ctxt_t ctx=(code_ctxt_t)0; all_code[addr] && ctx < (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {
      abs_loc PC = abs_loc(addr,ctx);
      if (state(PC)->num_calls > 0) {
        if (state(PC)->wcet != NEG_INFINITY) {
          cout << "Cost of " << addr_to_head_label(PC.addr) 
               << " (" << PC << ")"
               << " = " 
               << dec << state(PC)->wcet << hex
               << endl;
        } else {
          cout << "Cost of " << addr_to_head_label(PC.addr) << " unknown." << endl;
        }
      }
    }        
  }
}

////////////////////////////////////////////////////////////////
// Costs of each interrupt enable/disable
////////////////////////////////////////////////////////////////

static 
bool is_end_of_irq_enabled(abs_loc PC) {
  if (safe_ints_enabled(PC)) return false;

  for(unsigned int i=0; i<state(PC)->out_edges.len; ++i) {
    struct branch_elt *b = &state(PC)->out_edges.list[i];
    if (!b->dead) {
      abs_loc target = fixup_link(b->target,PC.ctx);
      if (avr_is_call(target)
          || avr_is_ret(target)) { 
      } else if (safe_ints_enabled(target)) {
        return true;
      }
    }
  }
  return false;
}

// Compute time that each irq is disabled
void report_irq_disable_time(abs_loc end) {
  scrub_wcets();

  if (debug_wcet > 1) {
    cout << "Pushing irq enable instruction " << end << endl;
  }
  wcet_queue.push(wcet_elt(end,(cost_t)0));

  queue<abs_loc> boundary;
  wcet_elt q;
  while (wcet_queue.pop(&q)) {
    if (safe_ints_enabled(q.PC)) {
      state(q.PC)->wcet = max<int>(state(q.PC)->wcet,q.wcet);
      boundary.push(q.PC);
    } else {
      wcet1(q);
    }
  }

  for(queue<abs_loc>::iterator i = boundary.begin();
      i != boundary.end();
      ++i) {
    abs_loc start = *i;
    if (state(start)->wcet != NEG_INFINITY
        && all_code[end.addr]->exn_count != 0) {

      queue<srcloc> *locs1 = 0;
      queue<srcloc> *locs2 = 0;
      if (0
          && addr_to_srclocs(start.addr,&locs1)
          && addr_to_srclocs(start.addr,&locs2)) {
        cout << "Interrupts disabled from " << start << " [";
        for(queue<srcloc>::iterator j = locs1->begin(); j != locs1->end(); ++j) {
          cout << " " << *j;
        }
        cout << "] to " << end << "[";
        for(queue<srcloc>::iterator j = locs2->begin(); j != locs2->end(); ++j) {
          cout << " " << *j;
        }
        cout << "]";
      } else {
        cout << "Interrupts disabled from " 
             << addr_to_head_label(start.addr) << "/" << start
             << " to "  
             << addr_to_head_label(end.addr) << "/" << end;
      }
      cout << " for " 
           << dec << state(start)->wcet / all_code[start.addr]->exn_count << hex
           << " cycles."
           << " Executed " << dec << all_code[start.addr]->exn_count << hex
           << " times."
           << endl;
    }    
  }
}


void report_irq_disable_times(void) {
  // Trace all live code starting from all irq re-enable points
  for(code_addr_t addr = 0; addr < (code_addr_t) code_size; ++addr) {
    if (all_code[addr]==0) continue;
    for (code_ctxt_t ctx=(code_ctxt_t)0; ctx < (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {
      abs_loc PC = abs_loc(addr,ctx);
      if (is_end_of_irq_enabled(PC)) {
        report_irq_disable_time(PC);
      }
    }
  }      
}

////////////////////////////////////////////////////////////////
// Report costs of ranges specified in file
////////////////////////////////////////////////////////////////

void report_cost_range(const cost_range &r) {
  assert(all_code[r.start]);
  assert(all_code[r.end]);
  assert(all_code[r.start]->state_cnt == all_code[r.end]->state_cnt);

  for(unsigned j = 0; j < all_code[r.start]->state_cnt; ++j) {
    abs_loc start = abs_loc(r.start,j);
    abs_loc end   = abs_loc(r.end,  j);

    assert(all_code[start.addr]->exn_count == all_code[end.addr]->exn_count);

    if (debug_wcet > 2) {
      cout << "Contemplating cost of " << r.label 
           << "." << dec << j << hex
           << endl;
    }
    scrub_wcets();
    wcet_queue.push(wcet_elt(end,0));
    wcet_fixpoint();
    
    if ( state(start)->wcet != NEG_INFINITY 
         && state(end)->wcet != NEG_INFINITY ) {
      cost_t diff = ((*state(start)).wcet - (*state(end)).wcet) / all_code[end.addr]->exn_count;
      cout << "Cost of " << r.label << "." << dec << j << hex << " "
           << "(" << start << " to " << end << ")"
           << " is " << dec << diff << hex
           << endl;
    } else {
      cout << "Drat and blast: didn't establish wcet from " 
           << start << " to " << end 
           << endl;
    }
  }
}

void report_cost_ranges(void) {
  for(queue<cost_range>::iterator i = cost_ranges.begin();
      i != cost_ranges.end();
      ++i) {
    report_cost_range(*i);
  }
  cout << endl;
}

////////////////////////////////////////////////////////////////
// Main entry point
////////////////////////////////////////////////////////////////

void wcet_analysis(void) {
  find_loops();

  report_function_costs();
  // We don't do this because it doesn't generate useful info
  // - all the irq disable times are short
  //  report_irq_disable_times();
  report_cost_ranges();

}

////////////////////////////////////////////////////////////////
// End
////////////////////////////////////////////////////////////////

#endif
